"""
67800 - Probabilistic Methods in AI
Spring 2017/18
Problem Set 1 - Bayesian Networks - Programming Task (6)
"""

import numpy as np 
import matplotlib.pyplot as plt
import pickle as pkl
import time
from scipy.io import loadmat


def get_p_z1(z1_val):
    '''
    Computes the prior probability for variable z1 to take value z1_val.
    '''
    return bayes_net['prior_z1'][z1_val]


def get_p_z2(z2_val):
    '''
    Computes the prior probability for variable z2 to take value z2_val.
    '''
    return bayes_net['prior_z2'][z2_val]


def get_p_xk_cond_z1_z2(z1_val, z2_val, k):
    '''
    Computes the conditional likelihood of variable xk assumes value 1
    given z1 assumes value z1_val and z2 assumes value z2_val.
    '''
    return bayes_net['cond_likelihood'][(z1_val, z2_val)][0, k]


def get_p_x_cond_z1_z2(z1_val, z2_val):
    '''
    Done
    '''
    return bayes_net['cond_likelihood'][(z1_val, z2_val)]


def get_pixels_sampled_from_p_x_joint_z1_z2():
    '''
    Done
    '''
    p_z1 = np.zeros((len(disc_z1),))
    for i, z1_val in enumerate(disc_z1):
        p_z1[i] = get_p_z1(z1_val)
    z1_sample = np.random.choice(disc_z1, 1, p=p_z1)[0]

    p_z2 = np.zeros((len(disc_z2),))
    for j, z2_val in enumerate(disc_z2):
        p_z2[j] = get_p_z2(z2_val)
    z2_sample = np.random.choice(disc_z2, 1, p=p_z2)[0]

    return np.random.binomial(1, get_p_x_cond_z1_z2(z1_sample, z2_sample))


def get_conditional_expectation(data):
    '''
    Done
    '''
    mean_z1_z2 = np.zeros((data.shape[0], 2))
    z1_z2 = np.ones((data.shape[0], 2))
    marginal_log_likelihood = get_log_p_x(data)
    for i, z1_val in enumerate(disc_z1):
        for j, z2_val in enumerate(disc_z2):
            posterior = get_log_p_x_joint_z1_z2(data, z1_val, z2_val) - marginal_log_likelihood
            mean_z1_z2 += np.expand_dims(np.exp(posterior),1) * ([z1_val, z2_val] * z1_z2)

    return mean_z1_z2[:, 0], mean_z1_z2[:, 1]


def get_log_p_x_joint_z1_z2(x_val, z1_val, z2_val):
    '''
    Computes the log-joint probability of all the variables.
    '''

    eps = 1e-7
    p_x_cond_z1_z2 = get_p_x_cond_z1_z2(z1_val, z2_val) + eps
    return np.sum(x_val * np.log(p_x_cond_z1_z2) + (1-x_val) * np.log(1-p_x_cond_z1_z2), axis=1) \
        + np.log(get_p_z1(z1_val) + eps) + np.log(get_p_z2(z2_val) + eps)


def get_log_p_x(x_val):
    '''
    Computes a lower bound on the marginal log-likelihood of the observed data.
    '''
    num_examples = x_val.shape[0]
    log_p_x = np.zeros((num_examples, len(disc_z1)*len(disc_z2)))
    for i, z1_val in enumerate(disc_z1):
        for j, z2_val in enumerate(disc_z2):
            col_index = i * len(disc_z1) + j
            log_p_x [:, col_index] = get_log_p_x_joint_z1_z2(x_val, z1_val, z2_val)

    from scipy.misc import logsumexp
    log_p_x = logsumexp(log_p_x, axis=1)

    return log_p_x


def q_6d():
    '''
    Plots the pixel variables sampled from the joint distribution as 28 x 28 images.
    Your job is to implement get_pixels_sampled_from_p_x_joint_z1_z2.
    '''

    plt.figure()
    for i in range(5):
        plt.subplot(1, 5, i+1)
        plt.imshow(get_pixels_sampled_from_p_x_joint_z1_z2().reshape(28, 28), cmap='gray')
        plt.title('Sample: ' + str(i+1))
    plt.tight_layout()
    plt.savefig('a_6d', bbox_inches='tight')
    plt.show()
    plt.close()

    return


def q_6e():
    '''
    Plots the expected images for each latent configuration on a 2D grid.
    Your job is to implement get_p_x_cond_z1_z2.
    '''

    canvas = np.empty((28*len(disc_z1), 28*len(disc_z2)))
    for i, z1_val in enumerate(disc_z1):
        for j, z2_val in enumerate(disc_z2):
            canvas[(len(disc_z1)-i-1)*28:(len(disc_z2)-i)*28, j*28:(j+1)*28] = \
                get_p_x_cond_z1_z2(z1_val, z2_val).reshape(28, 28)

    plt.figure()
    plt.imshow(canvas, cmap='gray')
    plt.tight_layout()
    plt.savefig('a_6e', bbox_inches='tight')
    plt.show()
    plt.close()

    return


def q_6f():
    '''
    Loads the data and plots the histograms. Rest is TODO.
    Your job is to compute real_marginal_log_likelihood and corrupt_marginal_log_likelihood below.
    '''

    mat = loadmat('q_6f.mat')
    val_data = mat['val_x']
    test_data = mat['test_x']

    real_marginal_log_likelihood = []
    corrupt_marginal_log_likelihood = []

    '''
    Done
    '''
    marginal_log_likelihood = get_log_p_x(val_data)
    val_mean = np.mean(marginal_log_likelihood)
    val_std = np.std(marginal_log_likelihood)
    marginal_log_likelihood = get_log_p_x(test_data)
    for mll in marginal_log_likelihood:
        if mll>=val_mean-3*val_std and mll<=val_mean+3*val_std:
            real_marginal_log_likelihood.append(mll)
        else:
            corrupt_marginal_log_likelihood.append(mll)

    plot_histogram(real_marginal_log_likelihood, title='Histogram of marginal log-likelihood for real data',
             xlabel='marginal log-likelihood', savefile='a_6f_hist_real')

    plot_histogram(corrupt_marginal_log_likelihood, title='Histogram of marginal log-likelihood for corrupted data',
        xlabel='marginal log-likelihood', savefile='a_6f_hist_corrupt')

    plt.show()
    plt.close()
    return


def q_6g():
    '''
    Loads the data and plots a color coded clustering of the conditional expectations.
    Your job is to implement the get_conditional_expectation function
    '''

    mat = loadmat('q_6g.mat')
    data = mat['x']
    labels = mat['y']

    mean_z1, mean_z2 = get_conditional_expectation(data)

    plt.figure()
    plt.scatter(mean_z1, mean_z2, c=np.squeeze(labels))
    plt.colorbar()
    plt.grid()
    plt.savefig('a_6g', bbox_inches='tight')
    plt.show()
    plt.close()

    return


def load_model(model_file):
    '''
    Loads a default Bayesian network with latent variables (in this case, a variational autoencoder)
    '''

    with open(model_file+'.pkl', 'rb') as infile:
        cpts = pkl.load(infile, encoding='bytes')

    model = {}
    model['prior_z1'] = cpts[0]
    model['prior_z2'] = cpts[1]
    model['cond_likelihood'] = cpts[2]

    return model


def plot_histogram(data, title='histogram', xlabel='value', ylabel='frequency', savefile='hist'):
    '''
    Plots a histogram.
    '''

    plt.figure()
    plt.hist(data)
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    plt.title(title)
    plt.savefig(savefile, bbox_inches='tight')
    return


def main():
    global disc_z1, disc_z2
    n_disc_z = 25
    disc_z1 = np.linspace(-3, 3, n_disc_z)
    disc_z2 = np.linspace(-3, 3, n_disc_z)

    global bayes_net
    bayes_net = load_model('trained_mnist_model')

    '''
    TODO: Using the above Bayesian Network model, complete the following parts.
    '''
    start_time = time.time()
    q_6d()
    # q_6e()
    # q_6f()
    # q_6g()
    print('Took %s secs' % (time.time() - start_time))
    return

if __name__== '__main__':
    main()
